"""Nodes to randomly rattle and shake structures."""

from abc import ABC, abstractmethod
from ase import Atoms
from typing import Iterable, Callable, NewType, Self, Iterator
from dataclasses import dataclass
import numpy as np

from .filters import Filter


def rattle(structure: Atoms, sigma: float) -> Atoms:
    """Randomly displace positions with gaussian noise.

    Operates INPLACE."""
    if len(structure) == 1:
        raise ValueError("Can only rattle structures larger than one atom.")
    structure.rattle(stdev=sigma)
    return structure


def stretch(structure: Atoms, hydro: float, shear: float, minimum_strain=1e-3) -> Atoms:
    """Randomly stretch cell with uniform noise.

    Ensures at least `minimum_strain` strain to avoid structures very close to their original structures.
    These don't offer a lot of new information and can also confuse VASP's symmetry analyzer.

    Operates INPLACE."""
    def scale(r):
        # map [0, 1) to [-1, 1)
        r = 2 * r - 1
        # map [-1, 1) to [-(1-m), 1-m)
        r *= (1 - minimum_strain)
        # map  [-(1-m), 1-m) to [-1, -m] u [m, 1)
        r += np.sign(r) * minimum_strain
        return r

    strain = shear * (2 * np.random.rand(3, 3) - 1)
    strain = 0.5 * (strain + strain.T)  # symmetrize
    np.fill_diagonal(strain, 1 + hydro * (2 * np.random.rand(3) - 1))
    structure.set_cell(structure.cell.array @ strain, scale_atoms=True)
    return structure


class ModifyABC(ABC):
    """Apply some modification to a given structure."""
    def __call__(self, structure: Atoms) -> Atoms:
        if 'modification' not in structure.info:
            structure.info['modification'] = str(self)
        else:
            structure.info['modification'] += '+' + str(self)
        return structure

    @abstractmethod
    def __str__(self) -> str:
        pass

    def __add__(self, other: Self) -> 'Series':
        return Series((self, other))

Modify = Callable[[Atoms], Atoms] | ModifyABC

def apply_modifications(
        structures: Iterable[Atoms],
        modifications: Iterable[Modify],
        filters: Iterable[Filter] | Filter | None = None,
) -> Iterator[Atoms]:
    """Apply a list of modifications to each structure and yield the result of each modification separately."""
    if filters is None:
        filters = []
    if not isinstance(filters, Iterable):
        filters = [filters]
    modifications = list(modifications)

    for structure in structures:
        for mod in modifications:
            try:
                m = mod(structure.copy())
            except ValueError:
                continue
            if all(f(m) for f in filters):
                yield m


@dataclass(frozen=True)
class Rattle(ModifyABC):
    """Displace atoms by some absolute amount from a normal distribution."""
    sigma: float
    create_supercells: bool = False
    "Create minimal 2x2x2 super cells when applied to structures of only one atom."

    def __call__(self, structure: Atoms):
        if self.create_supercells and len(structure) == 1:
            structure = structure.repeat(2)
        structure = super().__call__(structure)
        return rattle(structure, self.sigma)

    def __str__(self):
        return f"rattle({self.sigma})"


@dataclass(frozen=True)
class Stretch(ModifyABC):
    """Apply random cell modification."""
    hydro: float
    shear: float
    minimum_strain: float = 1e-3

    def __call__(self, structure: Atoms):
        structure = super().__call__(structure)
        return stretch(structure, self.hydro, self.shear, self.minimum_strain)

    def __str__(self):
        return f"stretch(hydro={self.hydro}, shear={self.shear})"


@dataclass(frozen=True)
class Series(ModifyABC):
    """Apply some modifications in sequence."""
    modifications: tuple[Modify, ...]

    def __call__(self, structure: Atoms) -> Atoms:
        for mod in self.modifications:
            structure = mod(structure)
        return structure

    def __str__(self):
        return "+".join(str(mod) for mod in self.modifications)


@dataclass(frozen=True)
class RandomChoice(ModifyABC):
    """Apply either of two alternatives randomly."""
    choice_a: Modify
    choice_b: Modify
    chance: float

    def __call__(self, structure: Atoms) -> Atoms:
        if np.random.rand() > self.chance:
            return self.choice_a(structure)
        else:
            return self.choice_b(structure)

    def __str__(self):
        return str(self.choice_a) + "|" + str(self.choice_b)



'''Classes that filter structures according to some criteria.

The code in the other modules that uses them is set up such that simple
functions can always be passed as well and that the classes here are just for
convenience.'''

from collections import defaultdict
from dataclasses import dataclass
from itertools import combinations_with_replacement
from math import nan, inf
from typing import Callable

from ase import Atoms
from structuretoolkit import get_neighbors

Filter = Callable[[Atoms], bool]


@dataclass
class DistanceFilter:
    '''Filter structures that contain too close atoms.

    Setting a radius to NaN allows all bonds involving this atom.'''
    radii: dict[str, float]

    @staticmethod
    def _element_wise_dist(structure: Atoms) -> dict[str, float]:
        pair = defaultdict(lambda: inf)
        # on weird aspect ratios the neighbor searching code can allocate huge structures,
        # because it explicitly repeats the structure to create ghost atoms
        # since we only care about the presence of short distances between atoms and not the
        # real neighbor information, simply double the structure to make sure we see all bonds
        # and turn off PBC
        # this can miss neighbors in structures with highly acute lattice angles, but we'll live
        sr = structure.repeat(2)
        sr.pbc = [False, False, False]
        n = get_neighbors(sr, num_neighbors=len(structure), mode="ragged")
        for i, (I, D) in enumerate(zip(n.indices, n.distances)):
            for j, d in zip(I, D):
                ei, ej = sorted((sr.symbols[i], sr.symbols[j]))
                pair[ei, ej] = min(d, pair[ei, ej])
        return pair

    def __call__(self, structure: Atoms) -> bool:
        '''
        Return True if structure satifies minimum distance criteria.

        Args:
            structure (ase.Atoms): structure to check

        Returns:
            `False`: at least on bond is shorter than the sum of given cutoff radii of the respective elements
            `True`: all bonds are than the sum of given cutoff radii of the respective elements
        '''
        pair = self._element_wise_dist(structure)
        for ei, ej in combinations_with_replacement(structure.symbols.species(), 2):
            ei, ej = sorted((ei, ej))
            if pair[ei, ej] < self.radii.get(ei, nan) + self.radii.get(ej, nan):
                return False
        return True


@dataclass
class AspectFilter:
    '''Filters structures with high aspect ratios.'''
    maximum_aspect_ratio: float = 6

    def __call__(self, structure: Atoms) -> bool:
        '''Return True if structure's cell has an agreeable aspect ratio.

        Args:
            structure (ase.Atoms): structure to check

        Returns:
            `True`: lattice's aspect ratio is below or equal `:attr:`.maximum_aspect_ratio`.
            `False`: lattice's aspect ratio is above `:attr:`.maximum_aspect_ratio`.'''
        a, b, c = sorted(structure.cell.lengths())
        return c / a <= self.maximum_aspect_ratio


@dataclass
class VolumeFilter:
    '''Filters structures by volume.'''
    maximum_volume_per_atom: float

    def __call__(self, structure: Atoms) -> bool:
        '''Return True if structure's volume is within range.

        Args:
            structure (ase.Atoms): structure to check

        Returns:
            `True`: volume per atom is smaller or equal than `:attr:.maximum_volume_per_atom`.
            `False`: otherwise'''
        return structure.cell.volume / len(structure) <= self.maximum_volume_per_atom


'''
Convenience shorts to create ASE calculators to be used inside ASSYST.

Exists mostly to avoid passing around potentially large and unpickle-able calculator objects.
'''
from abc import ABC, abstractmethod
from dataclasses import dataclass, asdict
from functools import lru_cache
import os

from ase.calculators.calculator import Calculator
from ase.calculators.morse import MorsePotential

class AseCalculatorConfig(ABC):
    '''Base class to keep calculator configurations.'''
    @abstractmethod
    def get_calculator(self) -> Calculator:
        '''Return the actual calculator object.

        Returns:
            :class:`ase.calculators.calculator.Calculator`: the actually usable calculator'''
        pass

@dataclass(frozen=True, eq=True)
class Grace(AseCalculatorConfig):
    '''Universal Graph Atomic Cluster Expansion models.

    .. attention::
        This class needs additional dependencies!
        Install `tensorpotential` from `PyPI <https://pypi.org/project/tensorpotential/>`__.'''
    model: str = "GRACE-FS-OAM"

    @lru_cache(maxsize=1)
    def get_calculator(self) -> Calculator:
        from tensorpotential.calculator import grace_fm
        # disable tensorflow warnings noise
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
        return grace_fm(self.model)

@dataclass(frozen=True, eq=True)
class Morse(AseCalculatorConfig):
    """Morse potential for testing.  Parameters as in ASE."""
    epsilon: float = 1.0
    r0: float = 1.0
    rho0: float = 1.0

    def get_calculator(self) -> Calculator:
        return MorsePotential(**asdict(self))



'''Relaxation step of ASSYST.'''
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Literal, Iterable, Iterator

from .calculators import AseCalculatorConfig

from ase import Atoms
from ase.calculators.calculator import Calculator
from ase.calculators.singlepoint import SinglePointCalculator
from ase.constraints import FixAtoms, FixSymmetry
from ase.filters import FrechetCellFilter
from ase.optimize import LBFGS

import numpy as np

@dataclass(frozen=True, eq=True)
class Relax:
    '''Minimize energy with respect to internal positions.

    Also used as a base class for all other relaxation.'''
    max_steps: int = 100
    force_tolerance: float = 1e-3
    algorithm: Literal["LBFGS"] = "LBFGS"

    def apply_filter_and_constraints(self, structure: Atoms):
        '''Hook to allow subclasses to add filters and constraints.'''
        return structure

    def relax(self, structure: Atoms) -> Atoms:
        '''Relax a structure and return result.

        Structure must have a calculator attached.
        Returned structure will have a SinglePointCalculator with the final energy, forces and stresses attached.

        Args:
            structure (ase.Atoms): structure to relax

        Returns:
            :class:`ase.Atoms`: relaxed structure with attached single point calculator.'''
        calc = structure.calc
        structure = structure.copy()
        structure.calc = calc
        lbfgs = LBFGS(self.apply_filter_and_constraints(structure), logfile="/dev/null")
        lbfgs.run(fmax=self.force_tolerance, steps=self.max_steps)
        structure.calc = None
        structure.calc = SinglePointCalculator(
                structure,
                energy=calc.get_potential_energy(),
                forces=calc.get_forces(),
                stress=calc.get_stress()
        )
        structure.constraints.clear()
        return structure


@dataclass(frozen=True, eq=True)
class CellRelax(Relax):
    '''Minimize energy while keeping relative positions and volume constant.'''

    def apply_filter_and_constraints(self, structure: Atoms):
        structure.set_constraint(FixAtoms(np.ones(len(structure),dtype=bool)))
        return FrechetCellFilter(structure, constant_volume=True)


@dataclass(frozen=True, eq=True)
class VolumeRelax(Relax):
    '''Minimize energy while keeping relative positions and cell shape constant.'''
    pressure: float = 0.0

    def apply_filter_and_constraints(self, structure: Atoms):
        structure.set_constraint(FixAtoms(np.ones(len(structure),dtype=bool)))
        return FrechetCellFilter(structure, hydrostatic_strain=True, scalar_pressure=self.pressure)


@dataclass(frozen=True, eq=True)
class SymmetryRelax(Relax):
    ''' Minimize energy with respect to internal positions and cell, while keeping space group fixed.'''
    pressure: float = 0.0

    def apply_filter_and_constraints(self, structure: Atoms):
        structure.set_constraint(FixSymmetry(structure))
        return FrechetCellFilter(structure, scalar_pressure=self.pressure)


@dataclass(frozen=True, eq=True)
class FullRelax(Relax):
    ''' Minimize energy with respect to internal positions and cell without constraints.'''
    pressure: float = 0.0

    def apply_filter_and_constraints(self, structure: Atoms):
        return FrechetCellFilter(structure, scalar_pressure=self.pressure)


def relax(
        settings: Relax,
        calculator: AseCalculatorConfig | Calculator,
        structure: Iterable[Atoms]
) -> Iterator[Atoms]:
    '''Relax structures according the given relaxation settings.

    Output structures have the final energy and force attached as ase's SinglePointCalculator.

    Args:
        settings (Relax): the kind of relaxation to perform (position, volume, etc.)
        calculator (AseCalculatorConfig or ase.calculators.calculator.Calculator): the energy/force engine to use
        structure (iterable of ase.Atoms): the structures to minimize

    Yields:
        :class:`ase.Atoms`: the corresponding relaxed configuration to each input structure
    '''
    for s in structure:
        s = s.copy()
        if isinstance(calculator, AseCalculatorConfig):
            s.calc = calculator.get_calculator()
        else:
            s.calc = calculator
        yield settings.relax(s)



import unittest
from unittest.mock import patch, MagicMock
from ase import Atoms

from assyst.structures import Formulas, sample_space_groups


class TestFormulas(unittest.TestCase):

    def test_unary_range(self):
        f = Formulas.unary_range("Cu", 1, 4)
        self.assertEqual(len(f), 3, msg="Length of unary_range('Cu', 1, 4) should be 3")
        self.assertEqual(f[0], {"Cu": 1}, msg="First element should be {'Cu': 1}")
        self.assertEqual(f[1], {"Cu": 2}, msg="Second element should be {'Cu': 2}")
        self.assertEqual(f[2], {"Cu": 3}, msg="Third element should be {'Cu': 3}")
        self.assertEqual(f.elements, {"Cu"}, msg="Elements should be {'Cu'}")

    def test_addition(self):
        f1 = Formulas.unary_range("Cu", 1, 3)
        f2 = Formulas.unary_range("Cu", 3, 5)
        combined = f1 + f2
        self.assertIsInstance(combined, Formulas, msg="Result of addition should be a Formulas instance")
        self.assertEqual(len(combined), 4, msg="Combined length should be 4")
        self.assertEqual(combined[0], {"Cu": 1}, msg="First element after addition should be {'Cu': 1}")
        self.assertEqual(combined[-1], {"Cu": 4}, msg="Last element after addition should be {'Cu': 4}")

    def test_or_operator(self):
        cu = Formulas.unary_range("Cu", 1, 3)
        ag = Formulas.unary_range("Ag", 1, 3)
        result = cu | ag
        self.assertIsInstance(result, Formulas, msg="Result of | operation should be a Formulas instance")
        self.assertIn({"Cu": 1, "Ag": 1}, result, msg="Result should contain {'Cu': 1, 'Ag': 1}")
        self.assertIn({"Cu": 2, "Ag": 2}, result, msg="Result should contain {'Cu': 2, 'Ag': 2}")

        with self.assertRaises(AssertionError, msg="Should raise AssertionError for overlapping elements"):
            _ = cu | cu

    def test_mul_operator(self):
        cu = Formulas.unary_range("Cu", 1, 3)
        ag = Formulas.unary_range("Ag", 1, 3)
        result = cu * ag
        expected = [
            {"Cu": 1, "Ag": 1},
            {"Cu": 1, "Ag": 2},
            {"Cu": 2, "Ag": 1},
            {"Cu": 2, "Ag": 2}
        ]
        self.assertEqual(len(result), 4, msg="Outer product should contain 4 combinations")
        for r in expected:
            self.assertIn(r, result, msg=f"Expected combination {r} missing in result")

        with self.assertRaises(AssertionError, msg="Should raise AssertionError for overlapping elements"):
            _ = cu * cu

    def test_sequence_protocol(self):
        f = Formulas.unary_range("Cu", 1, 3)
        self.assertIsInstance(f[0], dict, msg="Items in Formulas should be dicts")
        self.assertEqual(len(f), 2, msg="Length of unary_range('Cu', 1, 3) should be 2")


class TestSampleSpaceGroups(unittest.TestCase):

    @patch("assyst.structures.pyxtal")
    def test_max_structures(self, mock_pyxtal):
        mock_atoms = MagicMock(spec=Atoms)
        mock_pyxtal.return_value = [{"atoms": mock_atoms}] * 5

        f = Formulas.unary_range("Cu", 1, 3)
        results = list(sample_space_groups(f, max_structures=3))

        self.assertEqual(len(results), 3, msg="Should not generate more than max_structures=3")

    @patch("assyst.structures.pyxtal")
    def test_pyxtal_called_once_per_composition(self, mock_pyxtal):
        mock_atoms = MagicMock(spec=Atoms)
        mock_pyxtal.return_value = [{"atoms": mock_atoms}]

        # Define 3 compositions: Cu1, Cu2, Cu3
        formulas = Formulas.unary_range("Cu", 1, 4)  # 3 compositions

        results = list(sample_space_groups(formulas, max_structures=10))

        # We should get 3 results (since 1 per composition)
        self.assertEqual(len(results), 3, msg="Expected one structure per composition")
        self.assertEqual(mock_pyxtal.call_count, 3, msg="pyxtal should be called once per composition")

        expected_calls = [
            (('Cu',), (1,)),
            (('Cu',), (2,)),
            (('Cu',), (3,))
        ]
        actual_calls = [call.args for call in mock_pyxtal.call_args_list]

        for expected, actual in zip(expected_calls, actual_calls):
            self.assertEqual(
                    # actual called includes spacegroups that we did not include in the mock
                    actual[1:], expected,
                    msg=f"Expected pyxtal to be called with atom counts {expected[1]}, got {actual[1]}"
        )

    @patch("assyst.structures.pyxtal")
    def test_min_atoms(self, mock_pyxtal):
        mock_atoms = MagicMock(spec=Atoms)
        mock_pyxtal.return_value = [{"atoms": mock_atoms}]

        formulas = Formulas.unary_range("Cu", 1, 10)
        results = list(sample_space_groups(formulas, min_atoms=5))

        with self.subTest("unary"):
            for call in mock_pyxtal.call_args_list:
                self.assertLessEqual(5, sum(call.args[2]),
                    "sample_space_groups tried to call pyxtal with more atoms than it should have."
                )

        mock_pyxtal.reset_mock()

        formulas = Formulas.unary_range("Cu", 10) * Formulas.unary_range("Ag", 10)
        results = list(sample_space_groups(formulas, min_atoms=5))

        with self.subTest("binary"):
            for call in mock_pyxtal.call_args_list:
                self.assertLessEqual(5, sum(call.args[2]),
                    "sample_space_groups tried to call pyxtal with more atoms than it should have."
                )


if __name__ == "__main__":
    unittest.main()

    @patch("assyst.structures.pyxtal")
    def test_max_atoms(self, mock_pyxtal):
        mock_atoms = MagicMock(spec=Atoms)
        mock_pyxtal.return_value = [{"atoms": mock_atoms}]

        formulas = Formulas.unary_range("Cu", 1, 10)
        results = list(sample_space_groups(formulas, max_atoms=5))

        with self.subTest("unary"):
            for call in mock_pyxtal.call_args_list:
                self.assertLessEqual(sum(call.args[2]), 5, 
                    "sample_space_groups tried to call pyxtal with more atoms than it should have."
                )

        mock_pyxtal.reset_mock()

        formulas = Formulas.unary_range("Cu", 10) * Formulas.unary_range("Ag", 10)
        results = list(sample_space_groups(formulas, max_atoms=5))

        with self.subTest("binary"):
            for call in mock_pyxtal.call_args_list:
                self.assertLessEqual(sum(call.args[2]), 5, 
                    "sample_space_groups tried to call pyxtal with more atoms than it should have."
                )


if __name__ == "__main__":
    unittest.main()


'''Crystal structure generation step of ASSYST.'''

from dataclasses import dataclass
from collections.abc import Sequence
from itertools import product
from warnings import catch_warnings
from typing import Self, Iterable, Iterator, Literal

from ase import Atoms
from structuretoolkit.build.random import pyxtal
from pyxtal.tolerance import Tol_matrix
from tqdm.auto import tqdm
import math

@dataclass(eq=True, frozen=True)
class Formulas(Sequence):
    '''Simple helper to generate lists of structure compositions.

    :func:`.sample_space_groups` is the intended consumer and expects an iterable of dictionaries, where each dictionary
    maps an element name to the number of atoms of this type in one structure.
    :class:`.Formulas` behaves as if it were such a tuple, but extends the basic python arithmetic operations to make
    building the list a bit simpler.

    The class can be initialized from any tuple of dictionaries.

    >>> el_manual = Formulas(({'Cu': 1}, {'Cu': 2}))

    :meth:`.unary_range` is a helper class method that initializes `Formulas` for a single element and takes the same
    arguments as the builtin `range`, except that it skips the zero.

    >>> el = Formulas.unary_range('Cu', 3)
    Formulas(atoms=({'Cu': 1}, {'Cu': 2}))
    >>> el == el_manual
    True

    Addition is overloaded to the addition of the underlying tuples.

    >>> Formulas.unary_range('Cu', 1, 5) == Formulas.unary_range('Cu', 1, 3) + Formulas.unary_range('Cu', 3, 5)

    The bitwise or operation is akin to the inner product

    >>> Formulas.unary_range('Cu', 3) | Formulas.unary_range('Ag', 3)
    Formulas(atoms=({'Cu': 1, 'Ag': 1}, {'Cu': 2, 'Ag': 2}))

    >>> Formulas.unary_range('Cu', 3) * Formulas.unary_range('Ag', 3)
    Formulas(atoms=({'Cu': 1, 'Ag': 1}, {'Cu': 2, 'Ag': 1}, {'Cu': 1, 'Ag': 2}, {'Cu': 2, 'Ag': 2}))
    '''
    atoms: tuple[dict[str, int], ...]

    @property
    def elements(self) -> set[str]:
        '''Set of elements present in elements.'''
        e = set()
        for s in self.atoms:
            e = e.union(s.keys())
        return e

    @classmethod
    def unary_range(cls, element: str, *range_args) -> Self:
        '''Creates '''
        return cls(tuple({element: i} for i in range(*range_args) if i > 0))

    def __add__(self, other: Self) -> Self:
        '''Extend underlying list of stoichiometries.'''
        return Formulas(self.atoms + other.atoms)

    def __or__(self, other: Self) -> Self:
        '''Inner product of underlying stoichiometries.

        Truncates to the length of the shortest of the two element sequences.
        Must not share elements with other.elements.'''
        assert self.elements.isdisjoint(other.elements), "Can only or stoichiometries of different elements!"
        s = ()
        for me, you in zip(self.atoms, other.atoms):
            s += (me | you,)
        return Formulas(s)

    def __mul__(self, other: Self) -> Self:
        '''Outer product of underlying stoichiometries.

        Must not share elements with other.elements.'''
        assert self.elements.isdisjoint(other.elements), "Can only multiply stoichiometries of different elements!"
        s = ()
        for me, you in product(self.atoms, other.atoms):
            s += (me | you,)
        return Formulas(s)

    # Sequence Impl'
    def __getitem__(self, index: int) -> dict[str, int]:
        return self.atoms[index]

    def __len__(self) -> int:
        return len(self.atoms)

def sample_space_groups(
        formulas: Formulas | Iterable[dict[str, int]],
        spacegroups: list[int] | tuple[int,...] | None = None,
        min_atoms: int =  1,
        max_atoms: int = 10,
        max_structures: int | None = None,
        dim: Literal[0, 1, 2, 3] = 3,
) -> Iterator[Atoms]:
    '''
    Create symmetric random structures.

    Args:
        formulas (Formulas or iterable of dicts from str to int): list of chemical formulas
        spacegroups (list of int): which space groups to generate
        max_atoms (int): do not generate structures larger than this
        max_structures (int): generate at most this many structures
        dim (one of 0, 1, 2, or 3): the dimensionality of the structures to generate; if lower than 3 the code generates
            samples no longer from space groups, but from the subperiodic layer, rod, or point groups.

    Yields:
        `Atoms`: random symmetric crystal structures
    '''

    if spacegroups is None:
        spacegroups = list(range(1,231))
    if max_structures is None:
        max_structures = math.inf

    yielded = 0
    with catch_warnings(category=UserWarning, action='ignore'):
        for stoich in (bar := tqdm(formulas)):
            elements, num_atoms = zip(*stoich.items())
            if not min_atoms <= sum(num_atoms) <= max_atoms:
                continue
            stoich_str = "".join(f"{s}{n}" for s, n in zip(elements, num_atoms))
            bar.set_description(stoich_str)
            for s in pyxtal(spacegroups, elements, num_atoms, dim=dim, tm=Tol_matrix(prototype='metallic')):
                yield s['atoms']
                yielded += 1
                if yielded >= max_structures:
                    return



import unittest
import numpy as np
from ase import Atoms
from assyst.filters import DistanceFilter

class TestDistanceFilter(unittest.TestCase):
    def test_element_wise_dist(self):
        """_element_wise_dist returns a dict with all element pair keys present in structure."""
        with self.subTest("unary"):
            structure = Atoms('Cu2', cell=[4, 4, 4], pbc=True, positions=[(0, 0, 0), (2.0, 0, 0)])
            pair = DistanceFilter._element_wise_dist(structure)
            self.assertIsInstance(pair, dict, msg="Return type should be dict")
            self.assertIn(('Cu', 'Cu'), pair, msg="Pair ('Cu', 'Cu') should be present in unary case")
        with self.subTest("binary"):
            structure = Atoms('CuAg', cell=[4, 4, 4], pbc=True, positions=[(0, 0, 0), (2.5, 0, 0)])
            pair = DistanceFilter._element_wise_dist(structure)
            self.assertIsInstance(pair, dict, msg="Return type should be dict")
            self.assertIn(('Cu', 'Cu'), pair, msg="Pair ('Cu', 'Cu') should be present in binary case")
            self.assertIn(('Ag', 'Cu'), pair, msg="Pair ('Ag', 'Cu') should be present in binary case")
            self.assertIn(('Ag', 'Ag'), pair, msg="Pair ('Ag', 'Ag') should be present in binary case")

    def test_element_wise_dist_values(self):
        """_element_wise_dist should return correct minimum pair distances."""
        structure = Atoms('Cu2', cell=[4, 4, 4], pbc=True, positions=[(0, 0, 0), (2.0, 0, 0)])
        pair = DistanceFilter._element_wise_dist(structure)
        self.assertGreater(pair[('Cu', 'Cu')], 0, msg="Distance should be greater than zero")
        self.assertEqual(pair[('Cu', 'Cu')], 2.0, msg="Distance between Cu atoms is 2.0 units")

    def test_call_method(self):
        """__call__ returns True if all atomic pair distances exceed the sum of radii."""
        # Cu-Cu with radius 0.9, sum = 1.8, so distance 2.0 is valid
        structure = Atoms('Cu2', cell=[4, 4, 4], pbc=True, positions=[(0, 0, 0), (2.0, 0, 0)])
        filter = DistanceFilter({'Cu': 0.9})
        self.assertTrue(filter(structure), msg="Should be True: d=2.0 > 2*0.9")

    def test_call_method_false(self):
        """__call__ returns False if any atomic pair is closer than the sum of radii."""
        # Cu-Cu with radius 1.1, sum = 2.2, so distance 2.0 is invalid
        structure = Atoms('Cu2', cell=[4, 4, 4], pbc=True, positions=[(0, 0, 0), (2.0, 0, 0)])
        filter = DistanceFilter({'Cu': 1.1})
        self.assertFalse(filter(structure), msg="Should be False: d=2.0 < 2*1.1")

    def test_call_method_nan_radii(self):
        """__call__ returns True when radii for the relevant atom are NaN."""
        structure = Atoms('Cu2', cell=[4, 4, 4], pbc=True, positions=[(0, 0, 0), (2.0, 0, 0)])
        filter = DistanceFilter({'Cu': np.nan})
        self.assertTrue(filter(structure), msg="Should be True: radii is NaN")

    def test_call_method_empty_radii(self):
        """__call__ returns True when radii dictionary is empty."""
        structure = Atoms('Cu2', cell=[4, 4, 4], pbc=True, positions=[(0, 0, 0), (2.0, 0, 0)])
        filter = DistanceFilter({})
        self.assertTrue(filter(structure), msg="Should be True: radii dict is empty")

    def test_call_method_multiple_elements(self):
        """__call__ works for multi-element structures."""
        # Cu-Ag with radii 1.3 + 1.5 = 2.8, so all d > 2.8 is valid
        structure = Atoms('CuAg2', cell=[10, 10, 10], pbc=True,
                          positions=[(0, 0, 0), (3.0, 0, 0), (6.0, 0, 0)])
        filter = DistanceFilter({'Cu': 1.3, 'Ag': 1.5})
        self.assertTrue(filter(structure), msg="Should be True: all pair distances > radii sums")

    def test_call_method_multiple_elements_false(self):
        """__call__ returns False if any heterogeneous pair is too close (Cu-Ag < r_Cu + r_Ag)."""
        # Cu-Ag d = 2.5, sum = 2.8
        structure = Atoms('CuAg2', cell=[10, 10, 10], pbc=True,
                          positions=[(0, 0, 0), (2.5, 0, 0), (5.5, 0, 0)])
        filter = DistanceFilter({'Cu': 1.3, 'Ag': 1.5})
        self.assertFalse(filter(structure), msg="Should be False: d=2.5 < 1.3+1.5=2.8")

    def test_call_method_periodic_boundary(self):
        """__call__ correctly handles periodic boundary cases."""
        # One atom at 0, one at nearly cell edge, minimum image distance is 0.5
        structure = Atoms('Cu2', cell=[2.0, 2.0, 2.0], pbc=True, positions=[(0, 0, 0), (1.5, 0, 0)])
        filter = DistanceFilter({'Cu': 0.1})
        self.assertTrue(filter(structure), msg="minimal image d=0.5 > 2*0.1, so True")
        filter = DistanceFilter({'Cu': 0.3})
        self.assertFalse(filter(structure), msg="minimal image d=0.5 < 2*0.3")

if __name__ == '__main__':
    unittest.main()
